"""App for managing character standing requests, made for Alliance Auth."""

# pylint: disable = invalid-name

default_app_config = "standingsrequests.apps.StandingsRequestsConfig"

__version__ = "2.0.0"
__title__ = "Standings Requests"



# list of entity type IDs used in unit tests

from standingsrequests.core.contact_types import ContactTypeId

CHARACTER_TYPE_ID = ContactTypeId.CHARACTER_AMARR

CORPORATION_TYPE_ID = 2

ALLIANCE_TYPE_ID = 16159

FACTION_CALDARI_STATE_TYPE_ID = 500001

FACTION_TYPE_ID = FACTION_CALDARI_STATE_TYPE_ID



from io import StringIO
from unittest.mock import patch

from django.core.management import call_command
from django.test import TestCase, override_settings
from django.utils.timezone import now

from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import add_character_to_user

from standingsrequests.models import StandingRequest

from .testdata.my_test_data import (
    TEST_STANDINGS_ALLIANCE_ID,
    create_contacts_set,
    create_entity,
    load_eve_entities,
)

PACKAGE_PATH = "standingsrequests.management.commands"
TEST_USER_NAME = "Peter Parker"
TEST_REQUIRED_SCOPE = "mind_reading.v1"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(
    "standingsrequests.core.app_config.STR_ALLIANCE_IDS",
    [str(TEST_STANDINGS_ALLIANCE_ID)],
)
@patch(
    "standingsrequests.models.SR_REQUIRED_SCOPES",
    {"Member": [TEST_REQUIRED_SCOPE], "Blue": [], "": []},
)
@patch(PACKAGE_PATH + ".standingsrequests_sync_blue_alts.get_input")
class TestSyncRequests(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = AuthUtils.create_member(TEST_USER_NAME)
        AuthUtils.add_permission_to_user_by_name(
            StandingRequest.REQUEST_PERMISSION_NAME, cls.user
        )
        load_eve_entities()

    def setUp(self):
        self.contacts_set = create_contacts_set()
        self.out = StringIO()

    def test_abort_if_input_is_not_y(self, mock_get_input):
        mock_get_input.return_value = "N"
        alt = create_entity(EveCharacter, 1010)
        add_character_to_user(self.user, alt, scopes=[TEST_REQUIRED_SCOPE])

        call_command("standingsrequests_sync_blue_alts", stdout=self.out)

        self.assertEqual(StandingRequest.objects.count(), 0)

    def test_creates_new_request_for_blue_alt(self, mock_get_input):
        mock_get_input.return_value = "Y"
        alt = create_entity(EveCharacter, 1010)
        add_character_to_user(self.user, alt, scopes=[TEST_REQUIRED_SCOPE])

        call_command("standingsrequests_sync_blue_alts", stdout=self.out)

        self.assertEqual(StandingRequest.objects.count(), 1)
        request = StandingRequest.objects.first()
        self.assertEqual(request.user, self.user)
        self.assertEqual(request.contact_id, 1010)
        self.assertEqual(request.is_effective, True)
        self.assertAlmostEqual((now() - request.request_date).seconds, 0, delta=30)
        self.assertAlmostEqual((now() - request.action_date).seconds, 0, delta=30)
        self.assertAlmostEqual((now() - request.effective_date).seconds, 0, delta=30)



"""API for current configuration from settings."""

from typing import Optional, Set

from eveuniverse.models import EveEntity

from allianceauth.eveonline.models import EveCharacter

from standingsrequests.app_settings import (
    SR_OPERATION_MODE,
    STANDINGS_API_CHARID,
    STR_ALLIANCE_IDS,
    STR_CORP_IDS,
)
from standingsrequests.constants import OperationMode


def operation_mode() -> OperationMode:
    """Return current operation mode."""
    return OperationMode(SR_OPERATION_MODE)


def owner_character() -> EveCharacter:
    """Return the configured standings character."""
    try:
        return EveCharacter.objects.get(character_id=STANDINGS_API_CHARID)
    except EveCharacter.DoesNotExist:
        return EveCharacter.objects.create_character(STANDINGS_API_CHARID)


def standings_source_entity() -> Optional[EveEntity]:
    """Returns the entity that all standings are fetched from.

    When in alliance mode and character has no alliance, then return None
    """
    character = owner_character()
    if operation_mode() is OperationMode.ALLIANCE:
        if character.alliance_id:
            obj, _ = EveEntity.objects.get_or_create_esi(id=character.alliance_id)
            return obj

        return None

    if operation_mode() is OperationMode.CORPORATION:
        obj, _ = EveEntity.objects.get_or_create_esi(id=character.corporation_id)
        return obj

    raise NotImplementedError()


def is_character_a_member(character: EveCharacter) -> bool:
    """Return True if the character is in the organization, False otherwise."""
    return (
        character.corporation_id in corporation_ids()
        or character.alliance_id in alliance_ids()
    )


def corporation_ids() -> Set[int]:
    """Return corporation IDs, which belong the configured organization."""
    result = {int(org_id) for org_id in list(STR_CORP_IDS)}
    return result


def alliance_ids() -> Set[int]:
    """Return alliance IDs, which belong to the configured organization."""
    result = {int(org_id) for org_id in list(STR_ALLIANCE_IDS)}
    return result



from django.contrib import messages
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.shortcuts import redirect
from django.utils.translation import gettext_lazy as _

from standingsrequests.tasks import update_all


@login_required
@staff_member_required
def admin_changeset_update_now(request):
    update_all.delay(user_pk=request.user.pk)
    messages.info(
        request,
        _(
            "Started updating contacts and affiliations. "
            "You will receive a notification when completed."
        ),
    )
    return redirect("admin:standingsrequests_contactset_changelist")



from django.urls import path

from .views import (
    admin,
    create_requests,
    effective_requests,
    manage_requests,
    standings,
)

app_name = "standingsrequests"

urlpatterns = [
    # index
    path("", create_requests.index_view, name="index"),
    # admin
    path(
        "admin_changeset_update_now/",
        admin.admin_changeset_update_now,
        name="admin_changeset_update_now",
    ),
    # create_requests
    path("create_requests", create_requests.create_requests, name="create_requests"),
    path(
        "request_characters",
        create_requests.request_characters,
        name="request_characters",
    ),
    path(
        "request_corporations",
        create_requests.request_corporations,
        name="request_corporations",
    ),
    path(
        "request_character_standing/<int:character_id>/",
        create_requests.request_character_standing,
        name="request_character_standing",
    ),
    path(
        "remove_character_standing/<int:character_id>/",
        create_requests.remove_character_standing,
        name="remove_character_standing",
    ),
    path(
        "request_corp_standing/<int:corporation_id>/",
        create_requests.request_corp_standing,
        name="request_corp_standing",
    ),
    path(
        "remove_corp_standing/<int:corporation_id>/",
        create_requests.remove_corp_standing,
        name="remove_corp_standing",
    ),
    path("manage/setuptoken/", create_requests.view_auth_page, name="view_auth_page"),
    path(
        "requester_add_scopes/",
        create_requests.view_requester_add_scopes,
        name="view_requester_add_scopes",
    ),
    # effective requests
    path(
        "view/requests/",
        effective_requests.effective_requests,
        name="effective_requests",
    ),
    path(
        "view/requests/data/",
        effective_requests.effective_requests_data,
        name="effective_requests_data",
    ),
    # manage requests
    path("manage/", manage_requests.manage_standings, name="manage"),
    path(
        "manage/requests/",
        manage_requests.manage_requests_list,
        name="manage_requests_list",
    ),  # Should always follow the path of the GET path above
    path(
        "manage/requests/<int:contact_id>/",
        manage_requests.manage_requests_write,
        name="manage_requests_write",
    ),
    path(
        "manage/revocations/",
        manage_requests.manage_revocations_list,
        name="manage_revocations_list",
    ),
    path(
        "manage/revocations/<int:contact_id>/",
        manage_requests.manage_revocations_write,
        name="manage_revocations_write",
    ),
    # Standings
    path("standings", standings.standings, name="standings"),
    path(
        "standings/characters/data",
        standings.character_standings_data,
        name="character_standings_data",
    ),
    path(
        "standings/characters/download",
        standings.download_pilot_standings,
        name="download_pilots",
    ),
    path(
        "standings/corporations/data",
        standings.corporation_standings_data,
        name="corporation_standings_data",
    ),
    path(
        "standings/alliances/data",
        standings.alliance_standings_data,
        name="alliance_standings_data",
    ),
]



from django.contrib.auth.decorators import login_required, permission_required
from django.http import HttpResponse, HttpResponseNotFound
from django.shortcuts import get_object_or_404, render
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveEntity

from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from standingsrequests import __title__
from standingsrequests.app_settings import SR_NOTIFICATIONS_ENABLED
from standingsrequests.constants import DATETIME_FORMAT_HTML
from standingsrequests.core import app_config
from standingsrequests.models import (
    RequestLogEntry,
    StandingRequest,
    StandingRevocation,
)

from ._common import add_common_context, compose_standing_requests_data

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_standings(request):
    context = {
        "organization": app_config.standings_source_entity(),
        "requests_count": StandingRequest.objects.pending_requests().count(),
        "revocations_count": StandingRevocation.objects.pending_requests().count(),
    }
    return render(
        request,
        "standingsrequests/manage_requests.html",
        add_common_context(request, context),
    )


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_requests_list(request):
    requests_qs = StandingRequest.objects.pending_requests()
    requests_data = compose_standing_requests_data(requests_qs)
    context = {"DATETIME_FORMAT_HTML": DATETIME_FORMAT_HTML, "requests": requests_data}
    return render(
        request, "standingsrequests/partials/manage_requests_list.html", context
    )


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_revocations_list(request):
    revocations_qs = StandingRevocation.objects.pending_requests()
    revocations_data = compose_standing_requests_data(revocations_qs)
    context = {
        "DATETIME_FORMAT_HTML": DATETIME_FORMAT_HTML,
        "revocations": revocations_data,
    }
    return render(
        request, "standingsrequests/partials/manage_revocations_list.html", context
    )


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_requests_write(request, contact_id):
    contact_id = int(contact_id)
    logger.debug("manage_requests_write called by %s", request.user)
    if request.method == "PUT":
        actioned = 0
        for r in StandingRequest.objects.filter(contact_id=contact_id):
            r.mark_actioned(request.user)
            RequestLogEntry.objects.create_from_standing_request(
                r, RequestLogEntry.Action.CONFIRMED, request.user
            )
            actioned += 1
        if actioned > 0:
            return HttpResponse("")
        return HttpResponseNotFound()

    if request.method == "DELETE":
        standing_request = get_object_or_404(StandingRequest, contact_id=contact_id)
        RequestLogEntry.objects.create_from_standing_request(
            standing_request, RequestLogEntry.Action.REJECTED, request.user
        )
        standing_request.delete()
        if SR_NOTIFICATIONS_ENABLED:
            entity_name = EveEntity.objects.resolve_name(contact_id)
            title = _("Standing request for %s rejected") % entity_name
            message = _(
                "Your standing request for %(character)s has been rejected by %(user)s."
            ) % {
                "character": entity_name,
                "user": request.user,
            }

            notify(user=standing_request.user, title=title, message=message)
        return HttpResponse("")

    return HttpResponseNotFound()


@login_required
@permission_required("standingsrequests.affect_standings")
def manage_revocations_write(request, contact_id):
    contact_id = int(contact_id)
    logger.debug(
        "manage_revocations_write called by %s for contact_id %s",
        str(request.user),
        contact_id,
    )
    if request.method == "PUT":
        actioned = 0
        for r in StandingRevocation.objects.filter(
            contact_id=contact_id, action_date__isnull=True
        ):
            r.mark_actioned(request.user)
            RequestLogEntry.objects.create_from_standing_request(
                r, RequestLogEntry.Action.CONFIRMED, request.user
            )
            actioned += 1

        if actioned > 0:
            return HttpResponse("")

        return HttpResponseNotFound

    if request.method == "DELETE":
        standing_revocations_qs = StandingRevocation.objects.filter(
            contact_id=contact_id
        )
        standing_revocation = standing_revocations_qs.first()
        RequestLogEntry.objects.create_from_standing_request(
            standing_revocation, RequestLogEntry.Action.REJECTED, request.user
        )
        standing_revocations_qs.delete()
        if SR_NOTIFICATIONS_ENABLED and standing_revocation.user:
            entity_name = EveEntity.objects.resolve_name(contact_id)
            title = _("Standing revocation for %s rejected") % entity_name
            message = _(
                "Your standing revocation for %(character)s "
                "has been rejected by %(user)s."
            ) % {"character": entity_name, "user": request.user}

            notify(user=standing_revocation.user, title=title, message=message)
        return HttpResponse("")

    return HttpResponseNotFound()



# pylint: disable = redefined-builtin

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Tuple

from bravado.exception import HTTPError

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction
from django.db.models import Case, Q, Value, When
from django.utils.translation import gettext_lazy as _
from esi.models import Token
from eveuniverse.models import EveEntity
from eveuniverse.tasks import create_eve_entities

from allianceauth.eveonline.models import EveCharacter
from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.helpers import chunks
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import SR_NOTIFICATIONS_ENABLED
from .constants import CreateCharacterRequestResult, OperationMode
from .core import app_config
from .core.contact_types import ContactTypeId
from .providers import esi

if TYPE_CHECKING:
    from .models import AbstractStandingsRequest, ContactSet, StandingRequest

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class EsiContactsContainer:
    """Converts raw contacts and contact labels data from ESI into an object"""

    class EsiLabel:
        def __init__(self, json):
            self.id = json["label_id"]
            self.name = json["label_name"]

        def __str__(self) -> str:
            return str(self.name)

        def __repr__(self) -> str:
            return str(self)

    class EsiContact:
        def __init__(self, json, labels, names_info):
            self.id = json["contact_id"]
            self.name = names_info[self.id] if self.id in names_info else ""
            self.standing = json["standing"]
            self.in_watchlist = json["in_watchlist"] if "in_watchlist" in json else None
            self.label_ids = (
                json["label_ids"]
                if "label_ids" in json and json["label_ids"] is not None
                else []
            )
            # list of labels
            self.labels = [label for label in labels if label.id in self.label_ids]

        def __str__(self) -> str:
            return str(self.name)

        def __repr__(self):
            return str(self)

    def __init__(self, token, owner_character):
        self.contacts = []
        self.labels = []

        if app_config.operation_mode() is OperationMode.ALLIANCE:
            if not owner_character.alliance_id:
                raise RuntimeError(
                    "{owner_character}: owner character is not a member of an alliance"
                )
            labels = esi.client.Contacts.get_alliances_alliance_id_contacts_labels(
                alliance_id=owner_character.alliance_id,
                token=token.valid_access_token(),
            ).results()
            self.labels = [self.EsiLabel(label) for label in labels]
            contacts = esi.client.Contacts.get_alliances_alliance_id_contacts(
                alliance_id=owner_character.alliance_id,
                token=token.valid_access_token(),
            ).results()

        elif app_config.operation_mode() is OperationMode.CORPORATION:
            labels = (
                esi.client.Contacts.get_corporations_corporation_id_contacts_labels(
                    corporation_id=owner_character.corporation_id,
                    token=token.valid_access_token(),
                ).results()
            )
            self.labels = [self.EsiLabel(label) for label in labels]
            contacts = esi.client.Contacts.get_corporations_corporation_id_contacts(
                corporation_id=owner_character.corporation_id,
                token=token.valid_access_token(),
            ).results()
        else:
            raise NotImplementedError()

        logger.debug("Got %d contacts in total", len(contacts))
        entity_ids = [contact["contact_id"] for contact in contacts]
        resolver = EveEntity.objects.bulk_resolve_names(entity_ids)
        self.contacts = [
            self.EsiContact(contact, self.labels, resolver._names_map)
            for contact in contacts
        ]


class ContactSetManager(models.Manager):
    def create_new_from_api(self) -> Optional[ContactSet]:
        """fetches contacts with standings for configured alliance
        or corporation from ESI and stores them as newly created ContactSet

        Returns new ContactSet on success, else None
        """
        owner_character = app_config.owner_character()
        token: Token = (
            Token.objects.filter(character_id=owner_character.character_id)
            .require_scopes(self.model.required_esi_scope())
            .require_valid()
            .first()
        )
        if not token:
            logger.warning("Token for standing char could not be found")
            return None

        try:
            contacts_wrap = EsiContactsContainer(token, owner_character)
        except HTTPError as ex:
            logger.exception(
                "APIError occurred while trying to query api server: %s", ex
            )
            return None

        with transaction.atomic():
            contacts_set = self.create()
            self._add_labels_from_api(contacts_set, contacts_wrap.labels)
            self._add_contacts_from_api(contacts_set, contacts_wrap.contacts)

        return contacts_set

    def _add_labels_from_api(self, contact_set: ContactSet, labels):
        """Add the list of labels to the given ContactSet

        contact_set: ContactSet instance
        labels: Label dictionary
        """
        from .models import ContactLabel

        contact_labels = [
            ContactLabel(label_id=label.id, name=label.name, contact_set=contact_set)
            for label in labels
        ]
        ContactLabel.objects.bulk_create(contact_labels, ignore_conflicts=True)

    def _add_contacts_from_api(self, contact_set, contacts):
        """Add all contacts to the given ContactSet
        Labels _MUST_ be added before adding contacts

        :param contact_set: Django ContactSet to add contacts to
        :param contacts: List of _ContactsWrapper.Contact to add
        """
        from .models import Contact

        for contact in contacts:
            eve_entity, _ = EveEntity.objects.get_or_create_esi(id=contact.id)
            obj = Contact.objects.create(
                contact_set=contact_set,
                eve_entity=eve_entity,
                standing=contact.standing,
            )
            flat_labels = [label.id for label in contact.labels]
            labels = contact_set.labels.filter(label_id__in=flat_labels)
            obj.labels.add(*labels)


class ContactQuerySet(models.QuerySet):
    def filter_characters(self):
        return self.filter(eve_entity__category=EveEntity.CATEGORY_CHARACTER)

    def filter_corporations(self):
        return self.filter(eve_entity__category=EveEntity.CATEGORY_CORPORATION)

    def filter_alliances(self):
        return self.filter(eve_entity__category=EveEntity.CATEGORY_ALLIANCE)


class AbstractStandingsRequestQuerySet(models.QuerySet):
    def annotate_is_pending(self) -> models.QuerySet:
        return self.annotate(
            is_pending_annotated=Case(
                When(Q(action_date__isnull=True) & Q(is_effective=False), then=True),
                default=Value(False),
                output_field=models.BooleanField(),
            )
        )

    def annotate_is_actioned(self) -> models.QuerySet:
        return self.annotate(
            is_actioned_annotated=Case(
                When(Q(action_date__isnull=False) & Q(is_effective=False), then=True),
                default=Value(False),
                output_field=models.BooleanField(),
            )
        )


class _AbstractStandingsRequestManagerBase(models.Manager):
    def filter_characters(self) -> models.QuerySet:
        return self.filter(contact_type_id__in=ContactTypeId.character_ids())

    def filter_corporations(self) -> models.QuerySet:
        return self.filter(contact_type_id=ContactTypeId.CORPORATION)

    def process_requests(self) -> None:
        """Process all the Standing requests/revocation objects"""
        from .models import AbstractStandingsRequest

        if self.model is AbstractStandingsRequest:
            raise TypeError("Can not be called from abstract objects")

        organization = app_config.standings_source_entity()
        organization_name = organization.name if organization else ""
        query: models.QuerySet[AbstractStandingsRequest] = self.all()
        for standing_request in query:
            contact = EveEntity.objects.get_or_create_esi(
                id=standing_request.contact_id
            )[0]
            is_currently_effective = standing_request.is_effective
            is_satisfied_standing = standing_request.evaluate_effective_standing()
            if is_satisfied_standing and not is_currently_effective:
                if SR_NOTIFICATIONS_ENABLED:
                    self._notify_user_about_standing_change(
                        organization_name=organization_name,
                        standing_request=standing_request,
                        contact=contact,
                    )

                # if this was a revocation the standing requests need to be remove
                # to indicate that this character no longer has standing
                if standing_request.is_standing_revocation:
                    self._remove_standing_request_after_revocation(standing_request)

            elif is_satisfied_standing:
                # Just catching all other contact types (corps/alliances)
                # that are set effective
                pass

            elif not is_satisfied_standing and is_currently_effective:
                # Effective standing no longer effective
                self._removing_effective_standing(standing_request)

            else:
                # Check the standing hasn't been set actioned
                # and not updated in game
                actioned_timeout = standing_request.check_actioned_timeout()
                if actioned_timeout is not None and actioned_timeout:
                    logger.info(
                        "Standing request for contact ID %d has timed out "
                        "and will be reset",
                        standing_request.contact_id,
                    )
                    if SR_NOTIFICATIONS_ENABLED:
                        self._notify_user_about_timed_out_request(
                            standing_request, contact, actioned_timeout
                        )

    def _notify_user_about_standing_change(
        self,
        organization_name: str,
        standing_request: AbstractStandingsRequest,
        contact: EveEntity,
    ):
        if standing_request.is_standing_request:
            notify(
                user=standing_request.user,
                title=_("%s: Standing with %s now in effect")
                % (__title__, contact.name),
                message=_(
                    "'%(organization_name)s' now has blue standing with "
                    "your alt %(contact_category)s '%(contact_name)s'. "
                    "Please also update the standing of "
                    "your %(contact_category)s accordingly."
                )
                % {
                    "organization_name": organization_name,
                    "contact_category": contact.category,
                    "contact_name": contact.name,
                },
            )
        elif standing_request.is_standing_revocation:
            if standing_request.user:
                notify(
                    user=standing_request.user,
                    title=f"{__title__}: Standing with {contact.name} revoked",
                    message=_(
                        "'%(organization_name)s' no longer has "
                        "standing with your "
                        "%(contact_category)s '%(contact_name)s'. "
                        "Please also update the standing of "
                        "your %(contact_category)s accordingly."
                    )
                    % {
                        "organization_name": organization_name,
                        "contact_category": contact.category,
                        "contact_name": contact.name,
                    },
                )

    def _removing_effective_standing(self, standing_request: AbstractStandingsRequest):
        from .models import StandingRevocation

        logger.info(
            "Standing for %d is marked as effective but is not "
            "satisfied in game. Deleting.",
            standing_request.contact_id,
        )
        standing_request.delete(reason=StandingRevocation.Reason.REVOKED_IN_GAME)

    def _remove_standing_request_after_revocation(self, standing_request):
        from .models import StandingRequest, StandingRevocation

        StandingRequest.objects.filter(contact_id=standing_request.contact_id).delete()
        StandingRevocation.objects.filter(
            contact_id=standing_request.contact_id
        ).delete()

    def _notify_user_about_timed_out_request(
        self,
        standing_request: AbstractStandingsRequest,
        contact: EveEntity,
        actioned_timeout,
    ):
        title = _("Standing Request for %s reset") % contact.name
        message = (
            _(
                "The standing request for %(contact_category)s "
                "'%(contact_name)s' from %(user_name)s "
                "has been reset as it did not appear in "
                "game before the timeout period expired."
            )
            % {
                "contact_category": contact.category,
                "contact_name": contact.name,
                "user_name": standing_request.user.username,
            },
        )

        # Notify standing manager
        notify(user=actioned_timeout, title=title, message=message)
        # Notify the user
        notify(user=standing_request.user, title=title, message=message)

    def has_pending_request(self, contact_id: int) -> bool:
        """Checks if a request is pending for the given contact_id

        contact_id: int contact_id to check the pending request for

        returns True if a request is already pending, False otherwise
        """
        return self.pending_requests().filter(contact_id=contact_id).exists()

    def pending_requests(self) -> models.QuerySet:
        """returns all pending requests for this class"""
        return self.filter(action_date__isnull=True, is_effective=False)


AbstractStandingsRequestManager = _AbstractStandingsRequestManagerBase.from_queryset(
    AbstractStandingsRequestQuerySet
)


class StandingRequestManager(AbstractStandingsRequestManager):
    def validate_requests(self) -> int:
        """Validate all StandingsRequests and check
        that the user requesting them has permission and has API keys
        associated with the character/corp.

        StandingRevocation are created for invalid standing requests

        returns the number of invalid requests
        """
        from .models import StandingRevocation

        logger.debug("Validating standings requests")
        invalid_count = 0
        for standing_request in self.all():
            logger.debug(
                "Checking request for contact_id %d", standing_request.contact_id
            )
            reason = StandingRevocation.Reason.NONE
            if not standing_request.user.has_perm(self.model.REQUEST_PERMISSION_NAME):
                logger.debug("Request is invalid, user does not have permission")
                reason = StandingRevocation.Reason.LOST_PERMISSION
                is_valid = False

            elif (
                standing_request.is_corporation
                and not self.model.can_request_corporation_standing(
                    standing_request.contact_id, standing_request.user
                )
            ):
                logger.debug("Request is invalid, not all corp API keys recorded.")
                reason = StandingRevocation.Reason.MISSING_CORP_TOKEN
                is_valid = False

            else:
                is_valid = True

            if not is_valid:
                logger.info(
                    "Standing request for contact_id %d no longer valid. "
                    "Creating revocation",
                    standing_request.contact_id,
                )
                StandingRevocation.objects.add_revocation(
                    contact_id=standing_request.contact_id,
                    contact_type=self.model.contact_id_2_type(
                        standing_request.contact_type_id
                    ),
                    user=standing_request.user,
                    reason=reason,
                )
                invalid_count += 1

        return invalid_count

    def create_character_request(
        self, user: User, character: EveCharacter
    ) -> CreateCharacterRequestResult:
        """Create new character standings request for user if possible."""
        from .models import ContactSet, RequestLogEntry, StandingRevocation

        try:
            if character.character_ownership.user != user:
                logger.warning(
                    "%s: User %s does not own character, forbidden", character, user
                )
                return CreateCharacterRequestResult.USER_IS_NOT_OWNER

        except ObjectDoesNotExist:
            return CreateCharacterRequestResult.USER_IS_NOT_OWNER

        try:
            contact_set = ContactSet.objects.latest()
        except ContactSet.DoesNotExist:
            logger.warning("Failed to get a contact set")
            return CreateCharacterRequestResult.UNKNOWN_ERROR
        character_id = character.character_id

        if self.has_pending_request(
            character_id
        ) or StandingRevocation.objects.has_pending_request(character_id):
            logger.warning("%s: Character already has a pending request", character)
            return CreateCharacterRequestResult.CHARACTER_HAS_REQUEST

        if not self.model.has_required_scopes_for_request(
            character=character, user=user
        ):
            logger.warning("%s: Character does not have the required scopes", character)
            return CreateCharacterRequestResult.CHARACTER_IS_MISSING_SCOPES

        sr = self.get_or_create_2(
            user=user,
            contact_id=character_id,
            contact_type=self.model.ContactType.CHARACTER,
        )
        if contact_set.contact_has_satisfied_standing(character_id):
            sr.mark_actioned(user=None, reason=sr.Reason.STANDING_IN_GAME)
            sr.mark_effective()
            RequestLogEntry.objects.create_from_standing_request(
                sr, RequestLogEntry.Action.CONFIRMED, None
            )

        return CreateCharacterRequestResult.NO_ERROR

    def create_corporation_request(self, user: User, corporation_id: int) -> bool:
        """Create new corporation standings request for user if possible."""
        from .models import StandingRevocation

        if self.has_pending_request(
            corporation_id
        ) or StandingRevocation.objects.has_pending_request(corporation_id):
            logger.warning(
                "Contact ID %d already has a pending request", corporation_id
            )
            return False
        if not self.model.can_request_corporation_standing(corporation_id, user):
            logger.warning(
                "User %s does not have enough keys for corpID %d, forbidden",
                user,
                corporation_id,
            )
            return False
        self.get_or_create_2(
            user=user,
            contact_id=corporation_id,
            contact_type=self.model.ContactType.CORPORATION,
        )
        return True

    def get_or_create_2(
        self, user: User, contact_id: int, contact_type: str
    ) -> StandingRequest:
        """Get or create a new standing request

        Params:
        - user: User the request and contact_id belongs to
        - contact_id: contact_id to request standings on
        - contact_type: type of this contact

        Returns the created StandingRequest instance
        """
        contact_type_id = self.model.contact_type_2_id(contact_type)
        instance, _ = self.get_or_create(
            contact_id=contact_id,
            contact_type_id=contact_type_id,
            defaults={"user": user},
        )
        return instance


class StandingRevocationManager(AbstractStandingsRequestManager):
    def add_revocation(
        self,
        contact_id: int,
        contact_type: str,
        user: Optional[User] = None,
        reason: Optional[AbstractStandingsRequest.Reason] = None,
    ) -> object:
        """Add a new standings revocation

        Params:
        - contact_id: contact_id to request standings on
        - contact_type_id: contact_type_id from AbstractContact concrete implementation
        - user: user making the request

        Returns the created StandingRevocation instance
        """
        from .models import AbstractStandingsRequest

        logger.debug(
            "Adding new standings revocation for contact %d type %s",
            contact_id,
            contact_type,
        )
        contact_type_id = AbstractStandingsRequest.contact_type_2_id(contact_type)
        if self.has_pending_request(contact_id):
            logger.debug(
                "Cannot add revocation for contact %d %s, pending revocation exists",
                contact_id,
                contact_type_id,
            )
            return None

        if not reason:
            reason = AbstractStandingsRequest.Reason.NONE

        instance = self.create(
            contact_id=contact_id,
            contact_type_id=contact_type_id,
            user=user,
            reason=AbstractStandingsRequest.Reason(reason),
        )
        return instance


class CharacterAffiliationManager(models.Manager):
    def update_evecharacter_relations(self) -> None:
        """Update links to eve character in auth if any"""

        eve_character_id_map = {
            obj["character_id"]: obj["id"]
            for obj in EveCharacter.objects.values("id", "character_id")
        }
        with transaction.atomic():
            affiliations = self.filter(character_id__in=eve_character_id_map.keys())
            for affiliation in affiliations:
                affiliation.eve_character_id = eve_character_id_map[
                    affiliation.character_id
                ]
            self.bulk_update(
                objs=affiliations, fields=["eve_character_id"], batch_size=500
            )

    def update_from_esi(self) -> None:
        """Update all character affiliations we have contacts or requests for."""
        character_ids = self._gather_character_ids()
        if character_ids:
            affiliations = self._fetch_characters_affiliation_from_esi(character_ids)
            if affiliations:
                self._store_affiliations(affiliations)

    def _gather_character_ids(self) -> list:
        from .models import ContactSet, StandingRequest, StandingRevocation

        try:
            contact_set = ContactSet.objects.latest()
        except ContactSet.DoesNotExist:
            logger.warning("Could not find a contact set")
            return []

        character_ids_contacts = set(
            contact_set.contacts.filter_characters()
            .values_list("eve_entity_id", flat=True)
            .distinct()
        )
        character_ids_requests = set(
            StandingRequest.objects.filter_characters()
            .values_list("contact_id", flat=True)
            .distinct()
        )
        character_ids_revocations = set(
            StandingRevocation.objects.filter_characters()
            .values_list("contact_id", flat=True)
            .distinct()
        )
        return list(
            character_ids_contacts | character_ids_requests | character_ids_revocations
        )

    def _fetch_characters_affiliation_from_esi(self, character_ids) -> list:
        chunk_size = 1000
        affiliations = []
        for character_ids_chunk in chunks(character_ids, chunk_size):
            try:
                response = esi.client.Character.post_characters_affiliation(
                    characters=character_ids_chunk
                ).results()
            except HTTPError:
                logger.exception("Could not fetch character affiliations from ESI")
                return []

            affiliations += response

        return affiliations

    def _store_affiliations(self, affiliations) -> None:
        affiliation_objects = []
        for affiliation in affiliations:
            character, _ = EveEntity.objects.get_or_create(
                id=affiliation["character_id"]
            )
            corporation, _ = EveEntity.objects.get_or_create(
                id=affiliation["corporation_id"]
            )

            if affiliation.get("alliance_id"):
                alliance, _ = EveEntity.objects.get_or_create(
                    id=affiliation["alliance_id"]
                )

            else:
                alliance = None

            if affiliation.get("faction_id"):
                faction, _ = EveEntity.objects.get_or_create(
                    id=affiliation["faction_id"]
                )
            else:
                faction = None

            affiliation_objects.append(
                self.model(
                    character=character,
                    corporation=corporation,
                    alliance=alliance,
                    faction=faction,
                )
            )

        with transaction.atomic():
            self.all().delete()
            self.bulk_create(affiliation_objects, batch_size=500)

        new_ids = set()
        for obj in affiliation_objects:
            new_ids |= obj.entity_ids()

        EveEntity.objects.bulk_resolve_ids(new_ids)


class CorporationDetailsManager(models.Manager):
    def corporation_ids_from_contacts(self) -> set:
        from .models import Contact

        contact_corporation_ids = set(
            Contact.objects.filter_corporations().values_list(
                "eve_entity_id", flat=True
            )
        )
        character_affiliation_corporation_ids = set(
            Contact.objects.filter_characters().values_list(
                "eve_entity__character_affiliation__corporation_id", flat=True
            )
        )
        all_ids = contact_corporation_ids | character_affiliation_corporation_ids
        all_ids.discard(None)
        return all_ids

    def update_or_create_from_esi(self, id: int) -> Tuple[Any, bool]:
        """Updates or create an obj from ESI"""
        logger.info("%s: Fetching corporation from ESI", id)
        data = esi.client.Corporation.get_corporations_corporation_id(
            corporation_id=id
        ).results()
        corporation = EveEntity.objects.get_or_create(id=id)[0]
        alliance = (
            EveEntity.objects.get_or_create(id=data["alliance_id"])[0]
            if data.get("alliance_id")
            else None
        )
        ceo_id = data["ceo_id"] if data["ceo_id"] and data["ceo_id"] > 1 else None
        ceo = EveEntity.objects.get_or_create(id=ceo_id)[0] if ceo_id else None
        faction = (
            EveEntity.objects.get_or_create(id=data["faction_id"])[0]
            if data.get("faction_id")
            else None
        )
        EveEntity.objects.bulk_resolve_ids(
            filter(
                lambda x: x is not None,
                [id, data.get("alliance_id"), ceo_id, data.get("faction_id")],
            )
        )
        return self.update_or_create(
            corporation=corporation,
            defaults={
                "alliance": alliance,
                "ceo": ceo,
                "faction": faction,
                "member_count": data["member_count"],
                "ticker": data["ticker"],
            },
        )


class FrozenQuerySetMixin:
    """Ensures the update method can not be used."""

    def update(self, **kwargs) -> int:
        raise RuntimeError("Update not allowed for this model.")


class RequestLogEntryQuerySet(FrozenQuerySetMixin, models.QuerySet):
    pass


class RequestLogEntryManagerBase(models.Manager):
    # TODO: This method should be called as tasks, and entities should be resolved
    def create_from_standing_request(
        self, standing_request: AbstractStandingsRequest, action, action_by: User
    ) -> Optional[Any]:
        from .models import FrozenAlt, FrozenAuthUser, RequestLogEntry

        requested_for: FrozenAlt = (
            FrozenAlt.objects.get_or_create_from_standing_request(standing_request)[0]
        )
        if action_by:
            action_by_obj: Optional[FrozenAuthUser] = (
                FrozenAuthUser.objects.get_or_create_from_user(action_by)[0]
            )
        else:
            action_by_obj = None

        requested_by_obj: FrozenAuthUser = (
            FrozenAuthUser.objects.get_or_create_from_user(standing_request.user)[0]
        )
        request_type = RequestLogEntry.RequestType.from_standing_request(
            standing_request
        )
        new_obj = self.create(
            action=RequestLogEntry.Action(action),
            action_by=action_by_obj,
            request_type=request_type,
            requested_at=standing_request.request_date,
            requested_by=requested_by_obj,
            requested_for=requested_for,
            reason=standing_request.reason,
        )
        eve_entity_ids = requested_for.entity_ids() | requested_by_obj.entity_ids()
        if action_by_obj:
            eve_entity_ids |= action_by_obj.entity_ids()

        create_eve_entities.delay(list(eve_entity_ids))
        return new_obj


RequestLogEntryManager = RequestLogEntryManagerBase.from_queryset(
    RequestLogEntryQuerySet
)


class FrozenAuthUserQuerySet(FrozenQuerySetMixin, models.QuerySet):
    pass


class FrozenAuthUserManagerBase(models.Manager):
    def get_or_create_from_user(self, user: User) -> Tuple[Any, bool]:
        main_character = user.profile.main_character
        if main_character:
            character_id = main_character.character_id
            corporation_id = main_character.corporation_id
            alliance_id = main_character.alliance_id
            faction_id = main_character.faction_id

        else:
            character_id = corporation_id = alliance_id = faction_id = None

        try:
            obj = self.get(
                user_id=user.id,
                character_id=character_id,
                corporation_id=corporation_id,
                alliance_id=alliance_id,
                faction_id=faction_id,
            )
            return obj, False

        except self.model.DoesNotExist:
            alliance = (
                EveEntity.objects.get_or_create(id=alliance_id)[0]
                if alliance_id
                else None
            )
            character = (
                EveEntity.objects.get_or_create(id=character_id)[0]
                if character_id
                else None
            )
            corporation = (
                EveEntity.objects.get_or_create(id=corporation_id)[0]
                if corporation_id
                else None
            )
            faction = (
                EveEntity.objects.get_or_create(id=faction_id)[0]
                if faction_id
                else None
            )
            return self.get_or_create(
                user=user,
                character=character,
                corporation=corporation,
                alliance=alliance,
                faction=faction,
                state=user.profile.state,
            )


FrozenAuthUserManager = FrozenAuthUserManagerBase.from_queryset(FrozenAuthUserQuerySet)


class FrozenAltQuerySet(FrozenQuerySetMixin, models.QuerySet):
    pass


class FrozenAltManagerBase(models.Manager):
    def get_or_create_from_standing_request(
        self, standing_request: AbstractStandingsRequest
    ) -> Tuple[Any, bool]:
        eve_entity, _ = EveEntity.objects.get_or_create(id=standing_request.contact_id)

        if standing_request.is_character:
            category = self.model.Category.CHARACTER
            character = eve_entity
            try:
                alliance = character.character_affiliation.alliance
                corporation = character.character_affiliation.corporation
                faction = character.character_affiliation.faction
            except ObjectDoesNotExist:
                alliance = None
                corporation = None
                faction = None

        elif standing_request.is_corporation:
            category = self.model.Category.CORPORATION
            character = None
            corporation = eve_entity
            try:
                alliance = corporation.corporation_details.alliance
                faction = corporation.corporation_details.faction
            except ObjectDoesNotExist:
                alliance = None
                faction = None

        else:
            raise NotImplementedError()

        return self.get_or_create(
            character=character,
            corporation=corporation,
            alliance=alliance,
            category=category,
            faction=faction,
        )


FrozenAltManager = FrozenAltManagerBase.from_queryset(FrozenAltQuerySet)



import datetime as dt
from unittest.mock import patch

from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.utils.timezone import now
from eveuniverse.models import EveEntity

from allianceauth.eveonline.models import EveAllianceInfo, EveCharacter
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import add_character_to_user

from standingsrequests.core.contact_types import ContactTypeId
from standingsrequests.models import CharacterAffiliation, Contact, StandingRequest
from standingsrequests.tests.testdata.my_test_data import (
    create_contacts_set,
    create_eve_objects,
    load_corporation_details,
    load_eve_entities,
)
from standingsrequests.tests.utils import PartialDictEqualMixin, json_response_to_dict_2
from standingsrequests.views import standings
from standingsrequests.views.standings import _identify_main_for_character

TEST_SCOPE = "publicData"
MODULE_PATH = "standingsrequests.views.standings"


@patch("standingsrequests.core.app_config.STANDINGS_API_CHARID", 1001)
class TestStandingsView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eve_entities()
        create_eve_objects()
        cls.contact_set = create_contacts_set()
        CharacterAffiliation.objects.update_evecharacter_relations()

        cls.user = AuthUtils.create_member("John Doe")
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "standingsrequests.request_standings", cls.user
        )

    def test_can_open_standings_page(self):
        # given
        request = self.factory.get(reverse("standingsrequests:standings"))
        request.user = self.user
        # when
        response = standings.standings(request)
        # then
        self.assertEqual(response.status_code, 200)


class TestCharacterStandingsData(PartialDictEqualMixin, TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eve_entities()
        create_eve_objects()
        cls.contact_set = create_contacts_set()
        CharacterAffiliation.objects.update_evecharacter_relations()

        member_state = AuthUtils.get_member_state()
        member_state.member_alliances.add(EveAllianceInfo.objects.get(alliance_id=3001))
        cls.user = AuthUtils.create_member("John Doe")
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "standingsrequests.request_standings", cls.user
        )
        EveCharacter.objects.get(character_id=1009).delete()
        cls.main_character_1 = EveCharacter.objects.get(character_id=1002)
        cls.user_1 = AuthUtils.create_member(cls.main_character_1.character_name)
        add_character_to_user(
            cls.user_1,
            cls.main_character_1,
            is_main=True,
            scopes=[TEST_SCOPE],
        )
        cls.alt_character_1 = EveCharacter.objects.get(character_id=1007)
        add_character_to_user(
            cls.user_1,
            cls.alt_character_1,
            scopes=[TEST_SCOPE],
        )

    def test_normal_with_full_permissions(self):
        # given
        self.user = AuthUtils.add_permission_to_user_by_name(
            "standingsrequests.view", self.user
        )
        self.maxDiff = None
        request = self.factory.get(
            reverse("standingsrequests:character_standings_data")
        )
        request.user = self.user
        my_view_without_cache = standings.character_standings_data.__wrapped__
        # when
        response = my_view_without_cache(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict_2(response, "character_id")
        expected = {1001, 1002, 1003, 1004, 1005, 1006, 1008, 1009, 1010, 1110}
        self.assertSetEqual(set(data.keys()), expected)

        data_character_1002 = data[1002]
        expected = {
            "character_id": 1002,
            "corporation_name": "Wayne Technologies",
            "alliance_name": "Wayne Enterprises",
            "faction_name": "",
            "standing": 10.0,
            "labels_str": "blue, green",
            "main_character_name": "Peter Parker",
            "state": "Member",
        }
        self.assertPartialDictEqual(data_character_1002, expected)

        data_character_1009 = data[1009]
        expected = {
            "character_id": 1009,
            "corporation_name": "Lexcorp",
            "alliance_name": "",
            "faction_name": "",
            "standing": -10.0,
            "labels_str": "red",
            "main_character_name": "-",
            "state": "-",
        }
        self.assertPartialDictEqual(data_character_1009, expected)

    def test_normal_with_basic_permission(self):
        # given
        self.maxDiff = None
        request = self.factory.get(
            reverse("standingsrequests:character_standings_data")
        )
        request.user = self.user
        my_view_without_cache = standings.character_standings_data.__wrapped__
        # when
        response = my_view_without_cache(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict_2(response, "character_id")
        expected = {1001, 1002, 1003, 1004, 1005, 1006, 1008, 1009, 1010, 1110}
        self.assertSetEqual(set(data.keys()), expected)

        data_character_1002 = data[1002]
        expected = {
            "character_id": 1002,
            "corporation_name": "Wayne Technologies",
            "alliance_name": "Wayne Enterprises",
            "faction_name": "",
            "standing": 10.0,
            "labels_str": "blue, green",
            "main_character_name": "",
            "state": "",
        }
        self.assertPartialDictEqual(data_character_1002, expected)

    def test_identify_main_works_without_main(self):
        # given
        character = EveCharacter.objects.get(character_id=1004)
        add_character_to_user(
            self.user,
            character,
            scopes=[TEST_SCOPE],
        )
        character_entity = EveEntity.objects.get(id=1004)
        contact = Contact.objects.create(
            contact_set=self.contact_set,
            eve_entity=character_entity,
            standing=10.0,
        )
        # checks that there's no main defined
        self.assertIsNone(self.user.profile.main_character)
        # when
        state, main_character_name, main_character_html = _identify_main_for_character(
            contact
        )
        # then
        self.assertEqual(main_character_name, "No main associated")
        self.assertEqual(main_character_html, "")
        self.assertEqual(
            state, "Member"
        )  # AuthUtils.create_member gives them Member by default


class TestCorporationStandingsData(PartialDictEqualMixin, TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.contact_set = create_contacts_set()
        load_eve_entities()
        create_eve_objects()
        load_corporation_details()
        member_state = AuthUtils.get_member_state()
        member_state.member_alliances.add(EveAllianceInfo.objects.get(alliance_id=3001))
        cls.user_1 = AuthUtils.create_member("John Doe")
        cls.user_1 = AuthUtils.add_permission_to_user_by_name(
            "standingsrequests.request_standings", cls.user_1
        )
        EveCharacter.objects.get(character_id=1009).delete()
        cls.main_character_1 = EveCharacter.objects.get(character_id=1002)
        cls.user_2 = AuthUtils.create_member(cls.main_character_1.character_name)
        add_character_to_user(
            cls.user_2,
            cls.main_character_1,
            is_main=True,
            scopes=[TEST_SCOPE],
        )
        cls.alt_character_1 = EveCharacter.objects.get(character_id=1007)
        add_character_to_user(
            cls.user_2,
            cls.alt_character_1,
            scopes=[TEST_SCOPE],
        )
        StandingRequest.objects.create(
            user=cls.user_2,
            contact_id=2102,
            contact_type_id=ContactTypeId.CORPORATION,
            action_by=cls.user_1,
            action_date=now() - dt.timedelta(days=1, hours=1),
            is_effective=True,
            effective_date=now() - dt.timedelta(days=1),
        )

    def test_with_full_permissions(self):
        # given
        self.user_1 = AuthUtils.add_permission_to_user_by_name(
            "standingsrequests.view", self.user_1
        )
        self.maxDiff = None
        request = self.factory.get(
            reverse("standingsrequests:corporation_standings_data")
        )
        request.user = self.user_1
        my_view_without_cache = standings.corporation_standings_data.__wrapped__
        # when
        response = my_view_without_cache(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict_2(response, "corporation_id")
        self.assertSetEqual(set(data.keys()), {2001, 2003, 2102})
        obj = data[2001]
        expected = {
            "corporation_id": 2001,
            "alliance_name": "Wayne Enterprises",
            "faction_name": "",
            "standing": 10.0,
            "state": "-",
            "main_character_name": "-",
        }
        self.assertPartialDictEqual(obj, expected)
        obj = data[2102]
        self.assertPartialDictEqual(
            obj,
            {
                "corporation_id": 2102,
                "alliance_name": "",
                "faction_name": "",
                "standing": -10.0,
                "state": "Member",
                "main_character_name": "Peter Parker",
            },
        )

    def test_with_basic_permissions(self):
        # given
        self.maxDiff = None
        request = self.factory.get(
            reverse("standingsrequests:corporation_standings_data")
        )
        request.user = self.user_1
        my_view_without_cache = standings.corporation_standings_data.__wrapped__
        # when
        response = my_view_without_cache(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict_2(response, "corporation_id")
        obj = data[2102]
        self.assertPartialDictEqual(
            obj,
            {
                "corporation_id": 2102,
                "alliance_name": "",
                "faction_name": "",
                "standing": -10.0,
                "state": "",
                "main_character_name": "",
            },
        )


class TestAllianceStandingsData(PartialDictEqualMixin, TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.contact_set = create_contacts_set()
        load_eve_entities()
        create_eve_objects()
        load_corporation_details()
        member_state = AuthUtils.get_member_state()
        member_state.member_alliances.add(EveAllianceInfo.objects.get(alliance_id=3001))
        cls.user = AuthUtils.create_member("John Doe")
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "standingsrequests.request_standings", cls.user
        )
        EveCharacter.objects.get(character_id=1009).delete()
        cls.main_character_1 = EveCharacter.objects.get(character_id=1002)
        cls.user_1 = AuthUtils.create_member(cls.main_character_1.character_name)
        add_character_to_user(
            cls.user_1,
            cls.main_character_1,
            is_main=True,
            scopes=[TEST_SCOPE],
        )
        cls.alt_character_1 = EveCharacter.objects.get(character_id=1007)
        add_character_to_user(
            cls.user_1,
            cls.alt_character_1,
            scopes=[TEST_SCOPE],
        )

    def test_normal(self):
        # given
        self.maxDiff = None
        request = self.factory.get(reverse("standingsrequests:alliance_standings_data"))
        request.user = self.user
        my_view_without_cache = standings.alliance_standings_data.__wrapped__
        # when
        response = my_view_without_cache(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict_2(response, "alliance_id")
        self.assertSetEqual(set(data.keys()), {3010})
        obj = data[3010]
        self.assertPartialDictEqual(obj, {"alliance_id": 3010, "standing": -10.0})



import datetime as dt
from typing import Optional

from celery import Task, chain, shared_task

from django.contrib.auth.models import User
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import SR_STANDINGS_STALE_HOURS, SR_SYNC_BLUE_ALTS_ENABLED
from .core import app_config
from .models import (
    CharacterAffiliation,
    ContactSet,
    CorporationDetails,
    StandingRequest,
    StandingRevocation,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

TASK_DEFAULT_PRIORITY = 6
TASK_LOW_PRIORITY = 8


@shared_task(name="standings_requests.update_all", bind=True)
def update_all(self, user_pk: int = None):
    """Update all standings and affiliations."""
    priority = _determine_task_priority(self) or TASK_DEFAULT_PRIORITY
    tasks = [
        standings_update.si().set(priority=priority),
        update_associations_api.si().set(priority=priority),
    ]
    if user_pk:
        tasks.append(report_result_to_user.si(user_pk).set(priority=priority))

    chain(tasks).delay()


@shared_task(name="standings_requests.report_result_to_user")
def report_result_to_user(user_pk: int):
    try:
        user = User.objects.get(pk=user_pk)
    except User.DoesNotExist:
        logger.warning("Can not find a user with pk %d", user_pk)
        return

    source_entity = app_config.standings_source_entity()
    notify(
        user,
        _("%s: Standings loaded") % __title__,
        _("Standings have been successfully loaded for %s") % source_entity.name,
        level="success",
    )


@shared_task(name="standings_requests.standings_update", bind=True)
def standings_update(self):
    """Updates standings from ESI"""
    logger.info("Standings API update started")
    contact_set: Optional[ContactSet] = ContactSet.objects.create_new_from_api()
    if not contact_set:
        raise RuntimeError(
            "Standings API update returned None (API error probably),"
            "aborting standings update"
        )

    priority = _determine_task_priority(self) or TASK_DEFAULT_PRIORITY

    tasks = []

    if SR_SYNC_BLUE_ALTS_ENABLED:
        tasks.append(
            generate_standing_requests_for_blue_alts.si(contact_set.pk).set(
                priority=priority
            )
        )

    tasks.append(process_standing_requests.si().set(priority=priority))
    tasks.append(process_standing_revocations.si().set(priority=priority))

    chain(tasks).delay()


@shared_task
def generate_standing_requests_for_blue_alts(contact_set_pk: int):
    """Generate standing requests for blue alts."""
    contact_set = ContactSet.objects.get(pk=contact_set_pk)
    contact_set.generate_standing_requests_for_blue_alts()


@shared_task
def process_standing_requests():
    """Process standings requests."""
    StandingRequest.objects.process_requests()


@shared_task
def process_standing_revocations():
    """Process standing revocations."""
    StandingRevocation.objects.process_requests()


@shared_task(name="standings_requests.validate_requests")
def validate_requests():
    """Validate standings requests."""
    count = StandingRequest.objects.validate_requests()
    logger.info("Dealt with %d invalid standings requests", count)


@shared_task(name="standings_requests.update_associations_api", bind=True)
def update_associations_api(self):
    """Update character affiliations from ESI and relations to Eve Characters"""
    priority = _determine_task_priority(self) or TASK_DEFAULT_PRIORITY
    update_character_affiliations_from_esi.apply_async(priority=priority)
    update_character_affiliations_to_auth.apply_async(priority=priority)
    update_all_corporation_details.apply_async(priority=priority)


@shared_task
def update_character_affiliations_from_esi():
    CharacterAffiliation.objects.update_from_esi()
    logger.info("Finished character affiliations from ESI.")


@shared_task
def update_character_affiliations_to_auth():
    CharacterAffiliation.objects.update_evecharacter_relations()
    logger.info("Finished updating character affiliations to Auth.")


@shared_task(bind=True)
def update_all_corporation_details(self):
    existing_corporation_ids = (
        CorporationDetails.objects.corporation_ids_from_contacts()
    )
    CorporationDetails.objects.exclude(
        corporation_id__in=existing_corporation_ids
    ).delete()

    if not existing_corporation_ids:
        logger.info("No corporations to update.")
        return

    priority = _determine_task_priority(self) or TASK_DEFAULT_PRIORITY
    for corporation_id in existing_corporation_ids:
        update_corporation_detail.apply_async(args=[corporation_id], priority=priority)

    logger.info(
        "Started updating corporation details for %d corporations.",
        len(existing_corporation_ids),
    )


@shared_task
def update_corporation_detail(corporation_id: int):
    CorporationDetails.objects.update_or_create_from_esi(corporation_id)


@shared_task(name="standings_requests.purge_stale_data", bind=True)
def purge_stale_data(self):
    """Delete all stale contact sets, but always keep the newest."""
    try:
        latest_standings = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        logger.warning("No ContactSets available, nothing to delete")
        return

    cutoff_date = now() - dt.timedelta(hours=SR_STANDINGS_STALE_HOURS)
    stale_contacts_qs = ContactSet.objects.filter(date__lt=cutoff_date).exclude(
        id=latest_standings.id
    )
    stale_objs_count = stale_contacts_qs.count()
    if not stale_objs_count:
        logger.debug("No ContactSets to delete")
        return

    logger.info("Found %d stale contact sets to purge", stale_objs_count)
    priority = _determine_task_priority(self) or TASK_LOW_PRIORITY
    tasks = [
        purge_contact_set.si(obj.pk).set(priority=priority) for obj in stale_contacts_qs
    ]
    chain(tasks).delay()


@shared_task
def purge_contact_set(contact_set_pk: int):
    obj = ContactSet.objects.filter(pk=contact_set_pk)
    obj.delete()


def _determine_task_priority(task_obj: Task) -> Optional[int]:
    """Return priority of give task or None if not defined."""
    properties = task_obj.request.get("properties") or {}
    return properties.get("priority")



from django.apps import AppConfig

from . import __title__, __version__


class StandingsRequestsConfig(AppConfig):
    name = "standingsrequests"
    label = "standingsrequests"
    verbose_name = f"{__title__} v{__version__}"

    def ready(self):
        pass



from unittest.mock import patch

from django.test import TestCase
from eveuniverse.models import EveEntity

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import (
    NoSocketsTestCase,
    add_character_to_user,
    create_user_from_evecharacter,
)

from standingsrequests.helpers.evecorporation import EveCorporation
from standingsrequests.tests.testdata.my_test_data import (
    create_eve_objects,
    esi_get_corporations_corporation_id,
    get_my_test_data,
)

EVECORPORATION_PATH = "standingsrequests.helpers.evecorporation"
MODELS_PATH = "standingsrequests.models"


@patch(EVECORPORATION_PATH + ".cache")
@patch(EVECORPORATION_PATH + ".esi")
class TestEveCorporation(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.corporation = EveCorporation(
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            ticker="WYT",
            ceo_id=1003,
            member_count=3,
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
        )
        EveEntity.objects.create(id=3001, name="Wayne Enterprises", category="alliance")
        cls.maxDiff = None

    def test_init(self, mock_esi, mock_cache):
        self.assertEqual(self.corporation.corporation_id, 2001)
        self.assertEqual(self.corporation.corporation_name, "Wayne Technologies")
        self.assertEqual(self.corporation.ticker, "WYT")
        self.assertEqual(self.corporation.member_count, 3)
        self.assertEqual(self.corporation.alliance_id, 3001)
        self.assertEqual(self.corporation.alliance_name, "Wayne Enterprises")

    def test_str(self, mock_esi, mock_cache):
        expected = "Wayne Technologies"
        self.assertEqual(str(self.corporation), expected)

    def test_get_corp_by_id_not_in_cache(self, mock_esi, mock_cache):
        mock_Corporation = mock_esi.client.Corporation
        mock_Corporation.get_corporations_corporation_id.side_effect = (
            esi_get_corporations_corporation_id
        )
        expected = self.corporation
        mock_cache.get.return_value = None

        obj = EveCorporation.get_by_id(2001)
        self.assertEqual(obj, expected)
        self.assertTrue(mock_cache.set.called)

    def test_get_corp_by_id_not_in_cache_and_esi_failed(self, mock_esi, mock_cache):
        mock_Corporation = mock_esi.client.Corporation
        mock_Corporation.get_corporations_corporation_id.side_effect = (
            esi_get_corporations_corporation_id
        )
        mock_cache.get.return_value = None

        obj = EveCorporation.get_by_id(9876)
        self.assertIsNone(obj)

    def test_get_corp_by_id_in_cache(self, mock_esi, mock_cache):
        expected = self.corporation
        mock_cache.get.return_value = expected

        obj = EveCorporation.get_by_id(2001)
        self.assertEqual(obj, expected)

    def test_get_corp_esi(self, mock_esi, mock_cache):
        mock_esi.client.Corporation.get_corporations_corporation_id.side_effect = (
            esi_get_corporations_corporation_id
        )
        obj = EveCorporation.fetch_corporation_from_api(2102)
        self.assertEqual(obj.corporation_id, 2102)
        self.assertEqual(obj.corporation_name, "Lexcorp")
        self.assertEqual(obj.ticker, "LEX")
        self.assertEqual(obj.member_count, 2)
        self.assertIsNone(obj.alliance_id)

    def test_normal_corp_is_not_npc(self, mock_esi, mock_cache):
        normal_corp = EveCorporation(
            corporation_id=98397665,
            corporation_name="Rancid Rabid Rabis",
            ticker="RANCI",
            member_count=3,
            alliance_id=99005502,
            alliance_name="Same Great Taste",
        )
        self.assertFalse(normal_corp.is_npc)

    def test_npc_corp_is_npc(self, mock_esi, mock_cache):
        normal_corp = EveCorporation(
            corporation_id=1000134,
            corporation_name="Blood Raiders",
            ticker="TBR",
            member_count=22,
        )
        self.assertTrue(normal_corp.is_npc)

    def test_corp_without_members(self, mock_esi, mock_cache):
        normal_corp = EveCorporation(
            corporation_id=98397665,
            corporation_name="Rancid Rabid Rabis",
            ticker="RANCI",
        )
        self.assertIsNone(normal_corp.alliance_name)


class TestMemberTokensCountForUser(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        create_eve_objects()

    def test_should_count_valid_characters_only(self):
        # given
        user, _ = create_user_from_evecharacter(1001, scopes=["special-scope"])
        add_character_to_user(
            user, EveCharacter.objects.get(character_id=1002), scopes=["special-scope"]
        )  # same corp and valid scope
        add_character_to_user(
            user, EveCharacter.objects.get(character_id=1003)
        )  # same corp, but invalid scope
        add_character_to_user(
            user, EveCharacter.objects.get(character_id=1006)
        )  # different corp
        obj = EveCorporation(corporation_id=2001)

        # when
        with patch(MODELS_PATH + ".SR_REQUIRED_SCOPES", {"Guest": {"special-scope"}}):
            result = obj.member_tokens_count_for_user(user)

        # then
        self.assertEqual(result, 2)


class TestGetManyById(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        create_eve_objects()
        cls.corporations: dict = get_my_test_data()["EveCorporationInfo"]

    def test_should_return_corporations(self):
        def my_get_by_id(corporation_id, *args, **kwargs):
            try:
                obj = self.corporations[str(corporation_id)]
            except KeyError:
                return None

            return EveCorporation(**obj)

        # when
        with patch(
            EVECORPORATION_PATH + ".EveCorporation.get_by_id", new=my_get_by_id
        ), patch(EVECORPORATION_PATH + ".esi") as _:
            result = EveCorporation.get_many_by_id([2001, 2002, 2987])

        # then
        corporations = {obj.corporation_id: obj for obj in result}
        self.assertSetEqual(set(corporations.keys()), {2001, 2002})



from django.contrib.auth.decorators import login_required, permission_required
from django.db import models
from django.http import JsonResponse
from django.shortcuts import render
from django.utils.html import format_html
from django.views.decorators.cache import cache_page

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from standingsrequests import __title__
from standingsrequests.app_settings import SR_PAGE_CACHE_SECONDS
from standingsrequests.constants import DATETIME_FORMAT_PY
from standingsrequests.core import app_config
from standingsrequests.models import StandingRequest

from ._common import add_common_context, compose_standing_requests_data

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("standingsrequests.affect_standings")
def effective_requests(request):
    context = {
        "organization": app_config.standings_source_entity(),
        "requests_count": _standing_requests_to_view().count(),
    }
    return render(
        request,
        "standingsrequests/effective_requests.html",
        add_common_context(request, context),
    )


@cache_page(SR_PAGE_CACHE_SECONDS)
@login_required
@permission_required("standingsrequests.affect_standings")
def effective_requests_data(request):
    requests_data = compose_standing_requests_data(
        _standing_requests_to_view(), quick_check=True
    )
    for req in requests_data:
        req["request_date_str"] = {
            "display": req["request_date"].strftime(DATETIME_FORMAT_PY),
            "sort": req["request_date"].isoformat(),
        }
        req["labels_str"] = ", ".join(req["labels"])
        scopes_html = (
            '<i class="fas fa-check fa-fw text-success" title="Has required scopes"></i>'
            if req["has_scopes"]
            else '<i class="fas fa-times fa-fw text-danger" title="Does not have required scopes"></i>'
        )
        req["scopes_state_html"] = format_html(
            "{} {}", format_html(scopes_html), req["state"]
        )
        effective_html = (
            '<i class="fas fa-check fa-fw text-success" title="Standing Effective"></i>'
            if req["is_effective"]
            else '<i class="fas fa-times fa-fw text-danger" title="Standing not effective"></i>'
        )
        req["effective_html"] = format_html(
            "{} {}", format_html(effective_html), req["action_by"]
        )
        # remove columns that are not needed to reduce data volume
        del req["request_date"]
        del req["reason"]
        del req["main_character_ticker"]
        del req["main_character_icon_url"]
        del req["contact_icon_url"]
        del req["corporation_id"]
        del req["alliance_id"]
        del req["labels"]
        del req["is_effective"]
        del req["is_character"]
        del req["is_corporation"]
        del req["actioned"]
    return JsonResponse({"data": requests_data})


def _standing_requests_to_view() -> models.QuerySet:
    return (
        StandingRequest.objects.filter(is_effective=True)
        .select_related("user__profile")
        .order_by("-request_date")
    )



from typing import Set

from django.contrib.auth.models import User
from django.db import models
from eveuniverse.models import EveEntity


class FrozenModelMixin:
    """Objects of this model type can only be created, but not updated."""

    def save(self: models.Model, *args, **kwargs) -> None:
        if self.pk is None:
            super().save(*args, **kwargs)
        else:
            raise RuntimeError("No updates allowed for this object.")


class GatherEntityIdsMixin:
    """Add ability to gather all entity IDs from foreign keys of an object."""

    def entity_ids(self: models.Model) -> Set[int]:
        """Return all entity IDs in this object and ignore fields, which are None.

        The relevant fields are automatically detected.
        """
        relevant_fields = (
            field
            for field in self._meta.get_fields()
            if field.is_relation and field.related_model is EveEntity
        )
        values = (field.value_from_object(self) for field in relevant_fields)
        result = {value for value in values if value is not None}
        return result


def get_or_create_sentinel_user() -> User:
    """Get or create the sentinel user."""
    return User.objects.get_or_create(username="deleted")[0]



from enum import IntEnum
from typing import Set


class ContactTypeId(IntEnum):
    CHARACTER_AMARR = 1373
    CHARACTER_KUNNI = 1374
    CHARACTER_CIVRE = 1375
    CHARACTER_DETEIS = 1376
    CHARACTER_GALLENTE = 1377
    CHARACTER_INTAKI = 1378
    CHARACTER_SEBIESTOR = 1379
    CHARACTER_BRUTOR = 1380
    CHARACTER_STATIC = 1381
    CHARACTER_MODIFIER = 1382
    CHARACTER_ACHURA = 1383
    CHARACTER_JIN_MEI = 1384
    CHARACTER_KHANID = 1385
    CHARACTER_VHEROKIOR = 1386
    CHARACTER_DRIFTER = 34574
    CORPORATION = 2

    @property
    def is_character(self) -> bool:
        return self.value in self.character_ids()

    @property
    def is_corporation(self) -> bool:
        return self.value in self.corporation_ids()

    @classmethod
    def character_id(cls) -> int:
        return cls.CHARACTER_AMARR

    @classmethod
    def character_ids(cls) -> Set[int]:
        return {
            cls.CHARACTER_AMARR,
            cls.CHARACTER_KUNNI,
            cls.CHARACTER_CIVRE,
            cls.CHARACTER_DETEIS,
            cls.CHARACTER_GALLENTE,
            cls.CHARACTER_INTAKI,
            cls.CHARACTER_SEBIESTOR,
            cls.CHARACTER_BRUTOR,
            cls.CHARACTER_STATIC,
            cls.CHARACTER_MODIFIER,
            cls.CHARACTER_ACHURA,
            cls.CHARACTER_JIN_MEI,
            cls.CHARACTER_KHANID,
            cls.CHARACTER_VHEROKIOR,
            cls.CHARACTER_DRIFTER,
        }

    @classmethod
    def corporation_ids(cls) -> Set[int]:
        return {cls.CORPORATION}



from unittest.mock import patch

from django.test import TestCase

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import NoSocketsTestCase

from standingsrequests.core import app_config
from standingsrequests.tests.testdata.my_test_data import (
    create_entity,
    create_eve_objects,
    load_eve_entities,
)

MODULE_PATH = "standingsrequests.core.app_config"


@patch(MODULE_PATH + ".STANDINGS_API_CHARID", 1001)
class TestBaseConfig(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eve_entities()

    def test_should_return_existing_character(self):
        # given
        character = create_entity(EveCharacter, 1001)
        # when
        owner_character = app_config.owner_character()
        # then
        self.assertEqual(character, owner_character)

    @patch(MODULE_PATH + ".EveCharacter.objects.create_character")
    def test_create_new_character_if_not_exists(self, mock_create_character):
        # given
        character = create_entity(EveCharacter, 1002)
        mock_create_character.return_value = character
        # when
        owner_character = app_config.owner_character()
        # then
        self.assertEqual(character, owner_character)

    @patch(MODULE_PATH + ".SR_OPERATION_MODE", "alliance")
    def test_should_return_alliance(self):
        # given
        create_entity(EveCharacter, 1001)
        # when
        result = app_config.standings_source_entity()
        # then
        self.assertEqual(result.id, 3001)

    @patch(MODULE_PATH + ".SR_OPERATION_MODE", "corporation")
    def test_should_return_corporation(self):
        # given
        create_entity(EveCharacter, 1001)
        # when
        result = app_config.standings_source_entity()
        # then
        self.assertEqual(result.id, 2001)


class TestConfiguredOrganizations(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        create_eve_objects()
        cls.character_1001 = EveCharacter.objects.get(character_id=1001)

    @patch(MODULE_PATH + ".STR_CORP_IDS", ["2001"])
    @patch(MODULE_PATH + ".STR_ALLIANCE_IDS", [])
    def test_pilot_in_organization_matches_corp(self):
        self.assertTrue(app_config.is_character_a_member(self.character_1001))

    @patch(MODULE_PATH + ".STR_CORP_IDS", [])
    @patch(MODULE_PATH + ".STR_ALLIANCE_IDS", ["3001"])
    def test_pilot_in_organization_matches_alliance(self):
        self.assertTrue(app_config.is_character_a_member(self.character_1001))

    @patch(MODULE_PATH + ".STR_CORP_IDS", [])
    @patch(MODULE_PATH + ".STR_ALLIANCE_IDS", [3101])
    def test_pilot_in_organization_doest_not_exist(self):
        self.assertFalse(app_config.is_character_a_member(self.character_1001))

    @patch(MODULE_PATH + ".STR_CORP_IDS", [])
    @patch(MODULE_PATH + ".STR_ALLIANCE_IDS", [])
    def test_pilot_in_organization_matches_none(self):
        self.assertFalse(app_config.is_character_a_member(self.character_1001))



from django.test import TestCase
from eveuniverse.models import EveEntity

from allianceauth.tests.auth_utils import AuthUtils

from standingsrequests.models import FrozenAuthUser
from standingsrequests.tests.test_models import TEST_USER_NAME
from standingsrequests.tests.testdata.my_test_data import load_eve_entities


class TestGatherEntityIds(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eve_entities()

    def test_should_gather_all_entity_ids(self):
        # given
        user = AuthUtils.create_member(TEST_USER_NAME)
        alliance = EveEntity.objects.get(id=3001)
        character = EveEntity.objects.get(id=1001)
        corporation = EveEntity.objects.get(id=2001)
        faction = EveEntity.objects.create(
            id=5001, name="Caldari", category=EveEntity.CATEGORY_FACTION
        )
        alt = FrozenAuthUser.objects.create(
            alliance=alliance,
            character=character,
            corporation=corporation,
            faction=faction,
            user=user,
        )

        # when
        result = alt.entity_ids()

        # then
        expected = {alliance.id, character.id, corporation.id, faction.id}
        self.assertSetEqual(result, expected)

    def test_should_ignore_none_values(self):
        # given
        user = AuthUtils.create_member(TEST_USER_NAME)
        alliance = EveEntity.objects.get(id=3001)
        character = EveEntity.objects.get(id=1001)
        corporation = EveEntity.objects.get(id=2001)
        alt = FrozenAuthUser.objects.create(
            alliance=alliance, character=character, corporation=corporation, user=user
        )

        # when
        result = alt.entity_ids()

        # then
        expected = {alliance.id, character.id, corporation.id}
        self.assertSetEqual(result, expected)



from django.test import TestCase

from standingsrequests.core.contact_types import ContactTypeId
from standingsrequests.tests.testdata.entity_type_ids import (
    CHARACTER_TYPE_ID,
    CORPORATION_TYPE_ID,
)


class TestContactType(TestCase):
    def test_get_contact_type(self):
        self.assertEqual(ContactTypeId.character_id(), CHARACTER_TYPE_ID)

    def test_is_character(self):
        self.assertTrue(ContactTypeId(CHARACTER_TYPE_ID).is_character)
        self.assertFalse(ContactTypeId(CORPORATION_TYPE_ID).is_character)

    def test_get_contact_type_2(self):
        self.assertEqual(ContactTypeId.CORPORATION, CORPORATION_TYPE_ID)

    def test_is_corporation(self):
        self.assertFalse(ContactTypeId(CHARACTER_TYPE_ID).is_corporation)
        self.assertTrue(ContactTypeId(CORPORATION_TYPE_ID).is_corporation)



import enum

from django.db.models import TextChoices

DATETIME_FORMAT_HTML = "Y-m-d H:i"
DATETIME_FORMAT_PY = "%Y-%m-%d %H:%M"


class OperationMode(TextChoices):
    ALLIANCE = "alliance"
    CORPORATION = "corporation"


class CreateCharacterRequestResult(enum.IntEnum):
    """A result from StandingsRequests.objects.create_character_request."""

    NO_ERROR = enum.auto()
    USER_IS_NOT_OWNER = enum.auto()
    CHARACTER_HAS_REQUEST = enum.auto()
    CHARACTER_IS_MISSING_SCOPES = enum.auto()
    UNKNOWN_ERROR = enum.auto()


DEFAULT_IMAGE_SIZE = 32



import inspect
import json
import os
from copy import deepcopy
from typing import Any

from bravado.exception import HTTPNotFound

from eveuniverse.models import EveEntity

from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from app_utils.esi_testing import BravadoOperationStub, BravadoResponseStub

from standingsrequests.managers import EsiContactsContainer
from standingsrequests.models import (
    CharacterAffiliation,
    Contact,
    ContactSet,
    CorporationDetails,
)

TEST_STANDINGS_API_CHARID = 1001
TEST_STANDINGS_API_CHARNAME = "Bruce Wayne"
TEST_STANDINGS_CORPORATION_ID = 2001
TEST_STANDINGS_CORPORATION_NAME = "Wayne Technologies"
TEST_STANDINGS_ALLIANCE_ID = 3001
TEST_STANDINGS_ALLIANCE_NAME = "Wayne Enterprises"

TEST_SCOPE = "publicData"


##########################
# internal functions


def _load_test_data():
    current_dir = os.path.dirname(
        os.path.abspath(inspect.getfile(inspect.currentframe()))
    )

    with open(current_dir + "/my_test_data.json", "r", encoding="utf-8") as f:
        my_test_data = json.load(f)

    return my_test_data


def _load_entities():
    entities = {}
    for character_id, character in _my_test_data["EveCharacter"].items():
        entities[int(character_id)] = character["character_name"]

    for corporation_id, corporation in _my_test_data["EveCorporationInfo"].items():
        entities[int(corporation_id)] = corporation["corporation_name"]

    for alliance_id, alliance in _my_test_data["EveAllianceInfo"].items():
        entities[int(alliance_id)] = alliance["alliance_name"]

    return entities


_my_test_data = _load_test_data()
_entities = _load_entities()


##########################
# common functions


def get_my_test_data() -> dict:
    """returns the raw test data dict"""
    return _my_test_data


def get_entity_name(entity_id: int):
    """returns name if entity is found, else None"""

    if int(entity_id) in _entities:
        return _entities[int(entity_id)]
    else:
        return None


def get_entity_names(eve_entity_ids: list) -> dict:
    """returns dict with {id: name} for found entities, else empty dict"""
    names_info = {}
    for id in eve_entity_ids:
        name = get_entity_name(id)
        if name:
            names_info[id] = name

    return names_info


def get_entity_data(EntityClass: type, entity_id: int) -> Any:
    if EntityClass not in [EveCharacter, EveCorporationInfo, EveAllianceInfo]:
        raise TypeError(f"Invalid entity_class: {EntityClass.__name__}")

    if str(entity_id) not in _my_test_data[EntityClass.__name__]:
        raise ValueError(
            f"not entity found in test data for that entity_id = {entity_id}"
        )

    return _my_test_data[EntityClass.__name__][str(entity_id)]


def create_entity(EntityClass: type, entity_id: int) -> Any:
    """creates an Eve entity from test data"""
    data = get_entity_data(EntityClass, entity_id)
    return EntityClass.objects.create(**data)


##########################
# esi emulation


def esi_post_universe_names(ids, *args, **kwargs) -> Any:
    entities = []
    for entity in _my_test_data["esi_post_universe_names"]:
        if entity["id"] in ids:
            entities.append(entity)

    return BravadoOperationStub(entities)


def esi_post_characters_affiliation(characters, *args, **kwargs) -> Any:
    result = []
    for assoc in _my_test_data["CharacterAffiliation"]:
        if assoc["character_id"] in characters:
            row = assoc.copy()
            result.append(row)

    return BravadoOperationStub(result)


def esi_get_corporations_corporation_id(corporation_id, *args, **kwargs) -> Any:
    result = []
    corporation_id = str(corporation_id)
    if corporation_id not in _my_test_data["EveCorporationInfo"]:
        raise HTTPNotFound(BravadoResponseStub(404, reason="Test Exception"))

    row = _my_test_data["EveCorporationInfo"][corporation_id]
    result = {
        "name": row["corporation_name"],
        "ticker": row["corporation_ticker"],
        "member_count": row["member_count"],
        "ceo_id": row["ceo_id"] if "ceo_id" in row else None,
    }
    if row["alliance_id"]:
        result["alliance_id"] = row["alliance_id"]

    return BravadoOperationStub(result)


def esi_get_alliances_alliance_id_contacts_labels(*args, **kwargs) -> Any:
    return BravadoOperationStub(deepcopy(_my_test_data["alliance_labels"]))


def esi_get_alliances_alliance_id_contacts(*args, **kwargs) -> Any:
    return BravadoOperationStub(deepcopy(_my_test_data["alliance_contacts"]))


##########################
# app specific functions


def create_standings_char():
    character, _ = EveCharacter.objects.get_or_create(
        character_id=TEST_STANDINGS_API_CHARID,
        defaults={
            "character_name": TEST_STANDINGS_API_CHARNAME,
            "corporation_id": TEST_STANDINGS_CORPORATION_ID,
            "corporation_name": TEST_STANDINGS_CORPORATION_ID,
            "alliance_id": TEST_STANDINGS_ALLIANCE_ID,
            "alliance_name": TEST_STANDINGS_ALLIANCE_NAME,
        },
    )
    return character


def get_test_labels() -> list:
    """returns labels from test data as list of _ContactsWrapper.Label"""
    labels = []
    for label_data in get_my_test_data()["alliance_labels"]:
        labels.append(EsiContactsContainer.EsiLabel(label_data))

    return labels


def get_test_contacts():
    """returns contacts from test data as list of _ContactsWrapper.Contact"""
    labels = get_test_labels()

    contact_ids = [x["contact_id"] for x in get_my_test_data()["alliance_contacts"]]
    names_info = get_entity_names(contact_ids)
    contacts = []
    for contact_data in get_my_test_data()["alliance_contacts"]:
        contacts.append(
            EsiContactsContainer.EsiContact(contact_data, labels, names_info)
        )

    return contacts


def create_contacts_set(my_set: ContactSet = None, include_assoc=True) -> ContactSet:
    if not my_set:
        my_set = ContactSet.objects.create(name="Dummy Set")

    # add labels
    ContactSet.objects._add_labels_from_api(my_set, get_test_labels())

    # create contacts for ContactSet
    for contact in _my_test_data["alliance_contacts"]:
        if contact["contact_type"] == "character":
            category = EveEntity.CATEGORY_CHARACTER
        elif contact["contact_type"] == "corporation":
            category = EveEntity.CATEGORY_CORPORATION
        elif contact["contact_type"] == "alliance":
            category = EveEntity.CATEGORY_ALLIANCE
        else:
            raise ValueError("Invalid contact type")

        eve_entity, _ = EveEntity.objects.get_or_create(
            id=contact["contact_id"],
            defaults={
                "name": get_entity_name(contact["contact_id"]),
                "category": category,
            },
        )
        my_standing = Contact.objects.create(
            contact_set=my_set,
            eve_entity=eve_entity,
            standing=contact["standing"],
        )
        for label_id in contact["label_ids"]:
            my_standing.labels.add(my_set.labels.get(label_id=label_id))

    # update EveEntity based on characters
    for character_id, character_data in _my_test_data["EveCharacter"].items():
        EveEntity.objects.get_or_create(
            id=character_id,
            defaults={
                "name": character_data["character_name"],
                "category": EveEntity.CATEGORY_CHARACTER,
            },
        )
        EveEntity.objects.get_or_create(
            id=character_data["corporation_id"],
            defaults={
                "name": character_data["corporation_name"],
                "category": EveEntity.CATEGORY_CORPORATION,
            },
        )
        if character_data["alliance_id"]:
            EveEntity.objects.get_or_create(
                id=character_data["alliance_id"],
                defaults={
                    "name": character_data["alliance_name"],
                    "category": EveEntity.CATEGORY_ALLIANCE,
                },
            )

    # create CharacterAffiliation
    if include_assoc:
        CharacterAffiliation.objects.all().delete()
        for assoc in _my_test_data["CharacterAffiliation"]:
            CharacterAffiliation.objects.create(**assoc)

    return my_set


def create_eve_objects():
    """creates all Eve objects from test data"""
    for character_data in _my_test_data[EveCharacter.__name__].values():
        character = EveCharacter.objects.create(**character_data)
        if character.alliance_id:
            defaults = {
                "alliance_name": character.alliance_name,
                "alliance_ticker": character.alliance_ticker,
                "executor_corp_id": 2001,
            }
            alliance, _ = EveAllianceInfo.objects.get_or_create(
                alliance_id=character.alliance_id, defaults=defaults
            )
        else:
            alliance = None

        defaults = {
            "corporation_name": character.corporation_name,
            "corporation_ticker": character.corporation_ticker,
            "member_count": 99,
            "alliance": alliance,
        }
        EveCorporationInfo.objects.get_or_create(
            corporation_id=character.corporation_id, defaults=defaults
        )


def add_eve_object_to_eve_entities(obj: Any):
    if isinstance(obj, EveCharacter):
        EveEntity.objects.update_or_create(
            id=obj.character_id,
            defaults={
                "name": obj.character_name,
                "category": EveEntity.CATEGORY_CHARACTER,
            },
        )
        EveEntity.objects.update_or_create(
            id=obj.corporation_id,
            defaults={
                "name": obj.corporation_name,
                "category": EveEntity.CATEGORY_CORPORATION,
            },
        )
        if obj.alliance_id:
            EveEntity.objects.update_or_create(
                id=obj.alliance_id,
                defaults={
                    "name": obj.alliance_name,
                    "category": EveEntity.CATEGORY_ALLIANCE,
                },
            )
    elif isinstance(obj, EveCorporationInfo):
        EveEntity.objects.update_or_create(
            id=obj.corporation_id,
            defaults={
                "name": obj.corporation_name,
                "category": EveEntity.CATEGORY_CORPORATION,
            },
        )
    elif isinstance(obj, EveAllianceInfo):
        EveEntity.objects.update_or_create(
            id=obj.alliance_id,
            defaults={
                "name": obj.alliance_name,
                "category": EveEntity.CATEGORY_ALLIANCE,
            },
        )
    else:
        raise NotImplementedError()


def generate_eve_entities_from_allianceauth():
    for character in EveCharacter.objects.all():
        EveEntity.objects.get_or_create(
            id=character.character_id,
            defaults={
                "name": character.character_name,
                "category": EveEntity.CATEGORY_CHARACTER,
            },
        )
        EveEntity.objects.get_or_create(
            id=character.corporation_id,
            defaults={
                "name": character.corporation_name,
                "category": EveEntity.CATEGORY_CORPORATION,
            },
        )
        if character.alliance_id:
            EveEntity.objects.get_or_create(
                id=character.alliance_id,
                defaults={
                    "name": character.alliance_name,
                    "category": EveEntity.CATEGORY_ALLIANCE,
                },
            )
    for corporation in EveCorporationInfo.objects.all():
        EveEntity.objects.get_or_create(
            id=corporation.corporation_id,
            defaults={
                "name": corporation.corporation_name,
                "category": EveEntity.CATEGORY_CORPORATION,
            },
        )
    for alliance in EveAllianceInfo.objects.all():
        EveEntity.objects.get_or_create(
            id=alliance.alliance_id,
            defaults={
                "name": alliance.alliance_name,
                "category": EveEntity.CATEGORY_ALLIANCE,
            },
        )


def load_eve_entities():
    for character in _my_test_data["EveCharacter"].values():
        EveEntity.objects.get_or_create(
            id=character["character_id"],
            defaults={
                "name": character["character_name"],
                "category": EveEntity.CATEGORY_CHARACTER,
            },
        )
        EveEntity.objects.get_or_create(
            id=character["corporation_id"],
            defaults={
                "name": character["corporation_name"],
                "category": EveEntity.CATEGORY_CORPORATION,
            },
        )
        if character["alliance_id"]:
            EveEntity.objects.get_or_create(
                id=character["alliance_id"],
                defaults={
                    "name": character["alliance_name"],
                    "category": EveEntity.CATEGORY_ALLIANCE,
                },
            )
    for corporation in _my_test_data["EveCorporationInfo"].values():
        EveEntity.objects.get_or_create(
            id=corporation["corporation_id"],
            defaults={
                "name": corporation["corporation_name"],
                "category": EveEntity.CATEGORY_CORPORATION,
            },
        )
    for alliance in _my_test_data["EveAllianceInfo"].values():
        EveEntity.objects.get_or_create(
            id=alliance["alliance_id"],
            defaults={
                "name": alliance["alliance_name"],
                "category": EveEntity.CATEGORY_ALLIANCE,
            },
        )

    for id, name in [
        (500001, "Caldari State"),
        (500002, "Minmatar Republic"),
        (500003, "Amarr Empire"),
        (500004, "Gallente Federation"),
    ]:
        EveEntity.objects.get_or_create(
            id=id,
            defaults={"name": name, "category": EveEntity.CATEGORY_FACTION},
        )


def load_corporation_details():
    for record in _my_test_data["EveCorporationInfo"].values():
        alliance_id = record["alliance_id"] if record.get("alliance_id") else None
        CorporationDetails.objects.update_or_create(
            corporation_id=record["corporation_id"],
            defaults={
                "alliance_id": alliance_id,
                "ceo_id": 2102,
                "member_count": 99,
                "ticker": record["corporation_ticker"],
            },
        )



from dataclasses import dataclass
from typing import Dict, List, Optional

from django.contrib.auth.models import User
from django.db import models
from django.utils.html import format_html

from allianceauth.eveonline.models import EveCharacter

from standingsrequests import __title__
from standingsrequests.constants import DATETIME_FORMAT_HTML
from standingsrequests.core import app_config
from standingsrequests.core.contact_types import ContactTypeId
from standingsrequests.helpers.evecharacter import EveCharacterHelper
from standingsrequests.helpers.evecorporation import EveCorporation
from standingsrequests.models import (
    AbstractStandingsRequest,
    Contact,
    ContactSet,
    StandingRequest,
    StandingRevocation,
)

DEFAULT_ICON_SIZE = 32


def label_with_icon(icon_url: str, text: str):
    return format_html(
        '<span class="text-nowrap">'
        '<img src="{}" class="img-circle" style="width:{}px;height:{}px"> {}'
        "</span>",
        icon_url,
        DEFAULT_ICON_SIZE,
        DEFAULT_ICON_SIZE,
        text,
    )


def add_common_context(request, context: dict) -> dict:
    """adds the common context used by all view"""
    new_context = {
        **{
            "app_title": __title__,
            "operation_mode": str(app_config.operation_mode()),
            "pending_total_count": (
                StandingRequest.objects.pending_requests().count()
                + StandingRevocation.objects.pending_requests().count()
            ),
            "DATETIME_FORMAT_HTML": DATETIME_FORMAT_HTML,
        },
        **context,
    }
    return new_context


@dataclass(frozen=True)
class MainCharacterInfo:
    """A main character for output."""

    character_name: str = "-"
    ticker: str = "-"
    icon_url: str = "-"

    @property
    def is_valid(self) -> bool:
        return bool(self.character_name)

    def html(self) -> str:
        if not self.character_name:
            return ""

        return label_with_icon(self.icon_url, f"[{self.ticker}] {self.character_name}")

    @classmethod
    def create_from_user(cls, user: User) -> "MainCharacterInfo":
        if not user:
            return cls()

        main_character: EveCharacter = user.profile.main_character
        if not main_character:
            return cls()

        character_name = main_character.character_name
        ticker = main_character.corporation_ticker
        icon_url = main_character.portrait_url(DEFAULT_ICON_SIZE)
        obj = cls(character_name=character_name, ticker=ticker, icon_url=icon_url)
        return obj


@dataclass(frozen=True)
class OrganizationInfo:
    """Organizational info about a requestor."""

    contact_name: str = ""
    contact_icon_url: str = ""
    corporation_id: Optional[int] = None
    corporation_name: str = ""
    corporation_ticker: str = ""
    alliance_id: int = Optional[None]
    alliance_name: str = ""
    has_scopes: bool = False

    def contact_name_html(self) -> str:
        if not self.contact_name:
            return ""

        return label_with_icon(self.contact_icon_url, self.contact_name)

    def organization_html(self) -> str:
        if self.corporation_name:
            organization_html = f"[{self.corporation_ticker}] {self.corporation_name}"
            if self.alliance_name:
                organization_html = format_html(
                    "{}<br>{}", organization_html, self.alliance_name
                )
        else:
            organization_html = ""

        return organization_html

    @classmethod
    def create(
        cls,
        quick_check: bool,
        eve_characters: Dict[int, EveCharacter],
        eve_corporations: Dict[int, EveCorporation],
        req: AbstractStandingsRequest,
    ) -> "OrganizationInfo":
        if req.is_character:
            if req.contact_id in eve_characters:
                character = eve_characters[req.contact_id]
            else:
                # TODO: remove EveCharacterHelper usage
                character = EveCharacterHelper(req.contact_id)

            contact_name = character.character_name
            contact_icon_url = character.portrait_url(DEFAULT_ICON_SIZE)
            corporation_id = character.corporation_id
            corporation_name = (
                character.corporation_name if character.corporation_name else ""
            )
            corporation_ticker = (
                character.corporation_ticker if character.corporation_ticker else ""
            )
            alliance_id = character.alliance_id
            alliance_name = character.alliance_name if character.alliance_name else ""
            has_scopes = StandingRequest.has_required_scopes_for_request(
                character=character, user=req.user, quick_check=quick_check
            )
            return cls(
                contact_name,
                contact_icon_url,
                corporation_id,
                corporation_name,
                corporation_ticker,
                alliance_id,
                alliance_name,
                has_scopes,
            )

        if req.is_corporation and req.contact_id in eve_corporations:
            corporation = eve_corporations[req.contact_id]
            contact_icon_url = corporation.logo_url(DEFAULT_ICON_SIZE)
            contact_name = corporation.corporation_name
            corporation_id = corporation.corporation_id
            corporation_name = corporation.corporation_name
            corporation_ticker = corporation.ticker
            alliance_id = None
            alliance_name = ""
            has_scopes = (
                not corporation.is_npc
                and corporation.user_has_all_member_tokens(
                    user=req.user, quick_check=quick_check
                )
            )
            return cls(
                contact_name,
                contact_icon_url,
                corporation_id,
                corporation_name,
                corporation_ticker,
                alliance_id,
                alliance_name,
                has_scopes,
            )

        return cls()


def compose_standing_requests_data(
    requests_qs: models.QuerySet, quick_check: bool = False
) -> list:
    """composes list of standings requests or revocations based on queryset
    and returns it
    """
    requests_query: models.QuerySet[AbstractStandingsRequest] = (
        requests_qs.select_related(
            "user", "user__profile__state", "user__profile__main_character"
        )
    )
    eve_characters = _preload_eve_characters(requests_query)
    eve_corporations = _preload_eve_corporations(requests_query)
    contacts = _identify_contacts(eve_characters, eve_corporations)
    requests_data = []
    for req in requests_query:
        main_character = MainCharacterInfo.create_from_user(req.user)
        state_name = req.user.profile.state.name if req.user else "-"
        organization = OrganizationInfo.create(
            quick_check, eve_characters, eve_corporations, req
        )
        reason = req.get_reason_display() if req.is_standing_revocation else None
        requests_data.append(
            {
                "contact_id": req.contact_id,
                "contact_name": organization.contact_name,
                "contact_icon_url": organization.contact_icon_url,
                "contact_name_html": {
                    "display": organization.contact_name_html(),
                    "sort": organization.contact_name,
                },
                "corporation_id": organization.corporation_id,
                "corporation_name": organization.corporation_name,
                "corporation_ticker": organization.corporation_ticker,
                "alliance_id": organization.alliance_id,
                "alliance_name": organization.alliance_name,
                "organization_html": organization.organization_html(),
                "request_date": req.request_date,
                "action_date": req.action_date,
                "has_scopes": organization.has_scopes,
                "state": state_name,
                "reason": reason,
                "labels": sorted(_fetch_labels(contacts, req)),
                "main_character_name": main_character.character_name,
                "main_character_ticker": main_character.ticker,
                "main_character_icon_url": main_character.icon_url,
                "main_character_html": main_character.html(),
                "actioned": req.is_actioned,
                "is_effective": req.is_effective,
                "is_corporation": req.is_corporation,
                "is_character": req.is_character,
                "action_by": req.action_by.username if req.action_by else "(System)",
            }
        )
    return requests_data


# TODO: remove EveCorporation usage
def _preload_eve_corporations(
    requests_qs: models.QuerySet,
) -> Dict[int, EveCorporation]:
    corporation_ids = requests_qs.filter(
        contact_type_id=ContactTypeId.CORPORATION
    ).values_list("contact_id", flat=True)
    corporations = EveCorporation.get_many_by_id(corporation_ids)
    eve_corporations = {
        corporation.corporation_id: corporation for corporation in corporations
    }

    return eve_corporations


def _preload_eve_characters(requests_qs: models.QuerySet) -> Dict[int, EveCharacter]:
    eve_characters = EveCharacter.objects.filter(
        character_id__in=(
            requests_qs.exclude(contact_type_id=ContactTypeId.CORPORATION).values_list(
                "contact_id", flat=True
            )
        )
    )
    result = {character.character_id: character for character in eve_characters}

    return result


def _identify_contacts(eve_characters, eve_corporations) -> Dict[int, Contact]:
    try:
        contact_set = ContactSet.objects.latest()
    except ContactSet.DoesNotExist:
        contacts = {}
    else:
        all_contact_ids = set(eve_characters.keys()) | set(eve_corporations.keys())
        contacts = {
            obj.eve_entity_id: obj
            for obj in contact_set.contacts.prefetch_related("labels").filter(
                eve_entity_id__in=all_contact_ids
            )
        }
    return contacts


def _fetch_labels(
    contacts: Dict[int, Contact], req: AbstractStandingsRequest
) -> List[str]:
    try:
        my_contact = contacts[req.contact_id]
    except KeyError:
        labels = []
    else:
        labels = my_contact.labels_sorted
    return labels



import json
from datetime import timedelta
from typing import Any

from django.contrib.auth.models import User
from django.http import JsonResponse
from django.test import RequestFactory, TestCase
from django.utils.timezone import now

from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import add_character_to_user, response_text

from standingsrequests.core.contact_types import ContactTypeId
from standingsrequests.models import Contact, StandingRequest
from standingsrequests.tests.testdata.my_test_data import (
    TEST_SCOPE,
    create_contacts_set,
    create_eve_objects,
)


class PartialDictEqualMixin:
    def assertPartialDictEqual(self, d1: dict, d2: dict):
        """Assert that d1 equals d2 for the subset of keys of d1."""
        subset = {k: v for k, v in d1.items() if k in d2}
        self.assertDictEqual(subset, d2)


def json_response_to_python_2(response: JsonResponse, data_key="data") -> object:
    """Convert JSON response into Python object."""
    data = json.loads(response_text(response))
    return data[data_key]


def json_response_to_dict_2(response: JsonResponse, key="id", data_key="data") -> dict:
    """Convert JSON response into dict by given key."""
    return {x[key]: x for x in json_response_to_python_2(response, data_key)}


class TestViewPagesBase(PartialDictEqualMixin, TestCase):
    """Base TestClass for all tests that deal with standing requests

    Defines common test data
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()

        create_eve_objects()
        cls.contact_set = create_contacts_set()

        # State is alliance, all members can add standings
        member_state = AuthUtils.get_member_state()
        member_state.member_alliances.add(EveAllianceInfo.objects.get(alliance_id=3001))
        perm = AuthUtils.get_permission_by_name(StandingRequest.REQUEST_PERMISSION_NAME)
        member_state.permissions.add(perm)

        # Requesting user - can only make requests
        cls.main_character_1 = EveCharacter.objects.get(character_id=1002)
        cls.user_requestor = AuthUtils.create_member(
            cls.main_character_1.character_name
        )
        add_character_to_user(
            cls.user_requestor,
            cls.main_character_1,
            is_main=True,
            scopes=[TEST_SCOPE],
        )
        cls.alt_character_1 = EveCharacter.objects.get(character_id=1007)
        add_character_to_user(
            cls.user_requestor,
            cls.alt_character_1,
            scopes=[TEST_SCOPE],
        )
        cls.alt_corporation = EveCorporationInfo.objects.get(
            corporation_id=cls.alt_character_1.corporation_id
        )
        cls.alt_character_2 = EveCharacter.objects.get(character_id=1008)
        add_character_to_user(
            cls.user_requestor,
            cls.alt_character_2,
            scopes=[TEST_SCOPE],
        )

        # Standing manager - can do everything
        cls.main_character_2 = EveCharacter.objects.get(character_id=1001)
        cls.user_manager = AuthUtils.create_member(cls.main_character_2.character_name)
        add_character_to_user(
            cls.user_manager,
            cls.main_character_2,
            is_main=True,
            scopes=[TEST_SCOPE],
        )
        cls.user_manager = AuthUtils.add_permission_to_user_by_name(
            "standingsrequests.affect_standings", cls.user_manager
        )
        cls.user_manager = AuthUtils.add_permission_to_user_by_name(
            "standingsrequests.view", cls.user_manager
        )
        cls.user_manager = User.objects.get(pk=cls.user_manager.pk)

        # Old user - has no main and no rights
        cls.user_former_member = AuthUtils.create_user("Lex Luthor")
        cls.alt_character_3 = EveCharacter.objects.get(character_id=1010)
        add_character_to_user(
            cls.user_former_member,
            cls.alt_character_3,
            scopes=[TEST_SCOPE],
        )

    def _create_standing_for_alt(self, alt: Any) -> StandingRequest:
        if isinstance(alt, EveCharacter):
            contact_id = alt.character_id
            contact_type_id = ContactTypeId.character_id()
        elif isinstance(alt, EveCorporationInfo):
            contact_id = alt.corporation_id
            contact_type_id = ContactTypeId.CORPORATION
        else:
            raise NotImplementedError()

        return StandingRequest.objects.create(
            user=self.user_requestor,
            contact_id=contact_id,
            contact_type_id=contact_type_id,
            action_by=self.user_manager,
            action_date=now() - timedelta(days=1, hours=1),
            is_effective=True,
            effective_date=now() - timedelta(days=1),
        )

    def _set_standing_for_alt_in_game(self, alt: Any) -> None:
        if isinstance(alt, EveCharacter):
            contact_id = alt.character_id
            Contact.objects.update_or_create(
                contact_set=self.contact_set,
                eve_entity_id=contact_id,
                defaults={"standing": 10},
            )
        elif isinstance(alt, EveCorporationInfo):
            contact_id = alt.corporation_id
            Contact.objects.update_or_create(
                contact_set=self.contact_set,
                eve_entity_id=contact_id,
                defaults={"standing": 10},
            )
        else:
            raise NotImplementedError()

        self.contact_set.refresh_from_db()
